@@ -14,6 +14,20 @@ def _literal(node: ast.AST) -> Any:
     """Return a Python literal from an AST node or raise DSLParseError."""
     if isinstance(node, ast.Constant):
         return node.value
+    if isinstance(node, ast.JoinedStr):
+        parts: List[str] = []
+        for value in node.values:
+            if isinstance(value, ast.Constant):
+                parts.append(str(value.value))
+            elif isinstance(value, ast.FormattedValue):
+                try:
+                    expr = ast.unparse(value.value)  # type: ignore[attr-defined]
+                except Exception:
+                    expr = "?"
+                parts.append("{" + expr + "}")
+            else:
+                raise DSLParseError("Keyword argument values must be JSON-serialisable literals")
+        return "".join(parts)
     if isinstance(node, (ast.List, ast.Tuple)):
         return [_literal(elt) for elt in node.elts]
     if isinstance(node, ast.Dict):
@@ -62,7 +76,8 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
         versions: Dict[str, int] = {}
         latest: Dict[str, str] = {}
         context_suffix: str = ""
-        ctx_counts: Dict[str, int] = {"if": 0, "loop": 0, "while": 0}
+        ctx_counts: Dict[str, int] = {"if": 0, "loop": 0, "while": 0, "except": 0}
+        loop_depth = 0
 
         def _ssa_new(name: str) -> str:
             if not VALID_NAME_RE.match(name):
@@ -108,7 +123,7 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
             deps: List[str] = []
             for n in ast.walk(node):
                 if isinstance(n, ast.Name) and isinstance(n.ctx, ast.Load):
-                    if n.id not in callees and n.id not in deps:
+                    if n.id not in callees and n.id not in deps and n.id in latest:
                         deps.append(n.id)
             return deps
 
@@ -118,9 +133,10 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
             except Exception:
                 return node.__class__.__name__
 
-        def _emit_assign_from_call(var_name: str, call: ast.Call) -> str:
+        def _emit_assign_from_call(var_name: str, call: ast.Call, awaited: bool = False) -> str:
             op_name = _get_call_name(call.func)
             deps: List[str] = []
+            dep_labels: List[str] = []
 
             def _expand_star_name(ssa_var: str) -> List[str]:
                 # Expand if previous op was a PACK.*
@@ -129,27 +145,45 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
                         return list(prev.get("deps", []))
                 return [ssa_var]
 
-            for arg in call.args:
+            for idx, arg in enumerate(call.args):
                 if isinstance(arg, ast.Starred):
                     star_val = arg.value
                     if isinstance(star_val, ast.Name):
-                        deps.extend(_expand_star_name(_ssa_get(star_val.id)))
+                        expanded = _expand_star_name(_ssa_get(star_val.id))
+                        deps.extend(expanded)
+                        dep_labels.extend(["*"] * len(expanded))
                     elif isinstance(star_val, (ast.List, ast.Tuple)):
                         for elt in star_val.elts:
                             if not isinstance(elt, ast.Name):
                                 raise DSLParseError("Starred list/tuple elements must be names")
                             deps.append(_ssa_get(elt.id))
+                            dep_labels.append("*")
                     else:
                         raise DSLParseError("*args must be a name or list/tuple of names")
                 elif isinstance(arg, ast.Name):
                     deps.append(_ssa_get(arg.id))
+                    dep_labels.append("")
                 elif isinstance(arg, (ast.List, ast.Tuple)):
                     for elt in arg.elts:
                         if not isinstance(elt, ast.Name):
                             raise DSLParseError("List/Tuple positional args must be variable names")
                         deps.append(_ssa_get(elt.id))
+                        dep_labels.append("")
                 else:
-                    raise DSLParseError("Positional args must be variable names or lists/tuples of names")
+                    # Allow literal positional arguments by emitting a CONST node
+                    try:
+                        lit = _literal(arg)
+                    except DSLParseError:
+                        raise DSLParseError("Positional args must be variable names or lists/tuples of names or literals")
+                    const_id = _ssa_new(f"{var_name}_arg{idx}")
+                    ops.append({
+                        "id": const_id,
+                        "op": "CONST.value",
+                        "deps": [],
+                        "args": {"value": lit},
+                    })
+                    deps.append(const_id)
+                    dep_labels.append("")
 
             kwargs: Dict[str, Any] = {}
             for kw in call.keywords:
@@ -161,18 +195,39 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
                             kwargs[str(k)] = val
                     elif isinstance(v, ast.Name):
                         deps.append(_ssa_get(v.id))
+                        dep_labels.append("**")
                     else:
                         raise DSLParseError("**kwargs must be a dict literal or a variable name")
                 else:
                     if isinstance(kw.value, ast.Name):
                         deps.append(_ssa_get(kw.value.id))
+                        dep_labels.append(kw.arg or "")
                     else:
-                        kwargs[kw.arg] = _literal(kw.value)
+                        lit = _literal(kw.value)
+                        kwargs[kw.arg] = lit
+                        const_id = _ssa_new(f"{var_name}_{kw.arg}")
+                        ops.append(
+                            {
+                                "id": const_id,
+                                "op": "CONST.value",
+                                "deps": [],
+                                "args": {"value": lit},
+                            }
+                        )
+                        deps.append(const_id)
+                        dep_labels.append(kw.arg or "")
 
             ssa = _ssa_new(var_name)
-            ops.append({"id": ssa, "op": op_name, "deps": deps, "args": kwargs})
+            op: Dict[str, Any] = {"id": ssa, "op": op_name, "deps": deps, "args": kwargs, "dep_labels": dep_labels}
+            if awaited:
+                op["await"] = True
+            ops.append(op)
             return ssa
 
+        def _emit_expr_call(call: ast.Call, awaited: bool = False) -> str:
+            """Emit a node for a bare expression call (no assignment)."""
+            return _emit_assign_from_call("call", call, awaited)
+
         def _emit_assign_from_fstring(var_name: str, fstr: ast.JoinedStr) -> str:
             deps: List[str] = []
             parts: List[str] = []
@@ -222,6 +277,51 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
                         "args": {},
                     })
                     return ssa
+                if isinstance(value, ast.Dict):
+                    # Support dict with values from names/calls or literals by synthesizing nodes
+                    keys: List[str] = []
+                    deps: List[str] = []
+                    for k_node, v_node in zip(value.keys, value.values):
+                        k_str = _literal(k_node)
+                        if not isinstance(k_str, (str, int, float, bool)):
+                            k_str = str(k_str)
+                        keys.append(str(k_str))
+                        if isinstance(v_node, ast.Name):
+                            deps.append(_ssa_get(v_node.id))
+                        elif isinstance(v_node, ast.Await):
+                            inner = v_node.value
+                            if not isinstance(inner, ast.Call):
+                                raise DSLParseError("await must wrap a call in dict value")
+                            tmp_id = _emit_assign_from_call(f"{var_name}_field", inner, awaited=True)
+                            deps.append(tmp_id)
+                        elif isinstance(v_node, ast.Call):
+                            tmp_id = _emit_assign_from_call(f"{var_name}_field", v_node)
+                            deps.append(tmp_id)
+                        else:
+                            # Synthesize const for literal or arbitrary expression value
+                            try:
+                                lit_val = _literal(v_node)
+                            except DSLParseError:
+                                try:
+                                    lit_val = ast.unparse(v_node)  # type: ignore[attr-defined]
+                                except Exception:
+                                    lit_val = v_node.__class__.__name__
+                            tmp = _ssa_new(f"{var_name}_lit")
+                            ops.append({
+                                "id": tmp,
+                                "op": "CONST.value",
+                                "deps": [],
+                                "args": {"value": lit_val},
+                            })
+                            deps.append(tmp)
+                    ssa = _ssa_new(var_name)
+                    ops.append({
+                        "id": ssa,
+                        "op": "PACK.dict",
+                        "deps": deps,
+                        "args": {"keys": keys},
+                    })
+                    return ssa
                 raise
 
         def _emit_assign_from_comp(var_name: str, node: ast.AST) -> str:
@@ -245,6 +345,49 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
             })
             return ssa
 
+        def _emit_assign_from_subscript(var_name: str, node: ast.Subscript) -> str:
+            # Support name[key] where key is a JSON-serialisable literal
+            base = node.value
+            if not isinstance(base, ast.Name):
+                raise DSLParseError("Subscript base must be a variable name")
+            # Extract slice expression across Python versions
+            sl = getattr(node, 'slice', None)
+            # In Python >=3.9, slice is the actual node; before it may be ast.Index
+            if hasattr(ast, 'Index') and isinstance(sl, getattr(ast, 'Index')):  # type: ignore[attr-defined]
+                sl = sl.value  # type: ignore[assignment]
+            key = _literal(sl)  # may raise if not literal
+            ssa = _ssa_new(var_name)
+            ops.append({
+                "id": ssa,
+                "op": "GET.item",
+                "deps": [_ssa_get(base.id)],
+                "args": {"key": key},
+            })
+            return ssa
+
+        def _emit_assign_to_subscript(target: ast.Subscript, value: ast.AST) -> str:
+            """Emit a SET.item node for an assignment like ``name[key] = value``."""
+            base = target.value
+            if not isinstance(base, ast.Name):
+                raise DSLParseError("Subscript base must be a variable name")
+            # Extract slice expression across Python versions
+            sl = getattr(target, "slice", None)
+            if hasattr(ast, "Index") and isinstance(sl, getattr(ast, "Index")):  # type: ignore[attr-defined]
+                sl = sl.value  # type: ignore[assignment]
+            key = _literal(sl)
+
+            val_id = _emit_value(f"{base.id}_item", value)
+
+            base_id = _ssa_get(base.id)
+            ssa = _ssa_new(base.id)
+            ops.append({
+                "id": ssa,
+                "op": "SET.item",
+                "deps": [base_id, val_id],
+                "args": {"key": key},
+            })
+            return ssa
+
         def _emit_cond(node: ast.AST, kind: str = "if") -> str:
             expr = _stringify(node)
             deps = [_ssa_get(n) for n in _collect_value_deps(node)]
@@ -252,6 +395,51 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
             ops.append({"id": ssa, "op": "COND.eval", "deps": deps, "args": {"expr": expr, "kind": kind}})
             return ssa
 
+        def _emit_assign_from_ifexp(var_name: str, node: ast.IfExp) -> str:
+            """Emit operations for an inline ``a if cond else b`` expression."""
+            cond_id = _emit_cond(node.test, kind="ifexp")
+
+            then_start = len(ops)
+            then_id = _emit_value(f"{var_name}_then", node.body)
+            if len(ops) > then_start:
+                first = ops[then_start]
+                deps0 = first.get("deps", []) or []
+                if cond_id not in deps0:
+                    first["deps"] = [*deps0, cond_id]
+
+            else_start = len(ops)
+            else_id = _emit_value(f"{var_name}_else", node.orelse)
+            if len(ops) > else_start:
+                first = ops[else_start]
+                deps0 = first.get("deps", []) or []
+                if cond_id not in deps0:
+                    first["deps"] = [*deps0, cond_id]
+
+            ssa = _ssa_new(var_name)
+            ops.append({"id": ssa, "op": "PHI", "deps": [then_id, else_id], "args": {"var": var_name}})
+            return ssa
+
+        def _emit_value(var_name: str, value: ast.AST) -> str:
+            awaited = False
+            if isinstance(value, ast.Await):
+                value = value.value
+                awaited = True
+            if isinstance(value, ast.Call):
+                return _emit_assign_from_call(var_name, value, awaited)
+            if isinstance(value, ast.JoinedStr):
+                return _emit_assign_from_fstring(var_name, value)
+            if isinstance(value, (ast.Constant, ast.List, ast.Tuple, ast.Dict)):
+                return _emit_assign_from_literal_or_pack(var_name, value)
+            if isinstance(value, (ast.ListComp, ast.SetComp, ast.DictComp, ast.GeneratorExp)):
+                return _emit_assign_from_comp(var_name, value)
+            if isinstance(value, ast.Subscript):
+                return _emit_assign_from_subscript(var_name, value)
+            if isinstance(value, ast.IfExp):
+                return _emit_assign_from_ifexp(var_name, value)
+            if isinstance(value, ast.Name):
+                return _ssa_get(value.id)
+            raise DSLParseError("Right hand side must be a call or f-string")
+
         def _emit_iter(node: ast.AST, target_label: Optional[str] = None) -> str:
             expr = _stringify(node)
             deps = [_ssa_get(n) for n in _collect_value_deps(node)]
@@ -263,28 +451,24 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
             return ssa
 
         def _parse_stmt(stmt: ast.stmt) -> Optional[str]:
-            nonlocal returned_var, versions, latest, context_suffix
+            nonlocal returned_var, versions, latest, context_suffix, loop_depth
             if isinstance(stmt, ast.Assign):
-                if len(stmt.targets) != 1 or not isinstance(stmt.targets[0], ast.Name):
-                    raise DSLParseError("Assignment targets must be simple names")
-                var_name = stmt.targets[0].id
-                value = stmt.value
-                if isinstance(value, ast.Await):
-                    value = value.value
-                if isinstance(value, ast.Call):
-                    return _emit_assign_from_call(var_name, value)
-                elif isinstance(value, ast.JoinedStr):
-                    return _emit_assign_from_fstring(var_name, value)
-                elif isinstance(value, (ast.Constant, ast.List, ast.Tuple, ast.Dict)):
-                    return _emit_assign_from_literal_or_pack(var_name, value)
-                elif isinstance(value, (ast.ListComp, ast.SetComp, ast.DictComp, ast.GeneratorExp)):
-                    return _emit_assign_from_comp(var_name, value)
+                if len(stmt.targets) != 1:
+                    raise DSLParseError("Assignment must have exactly one target")
+                target = stmt.targets[0]
+                if isinstance(target, ast.Name):
+                    var_name = target.id
+                    return _emit_value(var_name, stmt.value)
+                elif isinstance(target, ast.Subscript):
+                    return _emit_assign_to_subscript(target, stmt.value)
                 else:
-                    raise DSLParseError("Right hand side must be a call or f-string")
+                    raise DSLParseError("Assignment targets must be simple names")
             elif isinstance(stmt, ast.Expr):
                 call = stmt.value
+                awaited = False
                 if isinstance(call, ast.Await):
                     call = call.value
+                    awaited = True
                 if not isinstance(call, ast.Call):
                     raise DSLParseError("Only call expressions allowed at top level")
                 name = _get_call_name(call.func)
@@ -310,21 +494,40 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
                         raise DSLParseError("output requires as=\"filename\"")
                     outputs.append({"from": ssa_from, "as": filename})
                 else:
-                    raise DSLParseError("Only settings() and output() calls allowed as expressions")
+                    # General expression call: represent as an op node too
+                    _emit_expr_call(call, awaited)
                 return None
             elif isinstance(stmt, ast.Return):
+                if loop_depth > 0:
+                    ssa = _ssa_new("break")
+                    ops.append({"id": ssa, "op": "CTRL.break", "deps": [], "args": {}})
                 if isinstance(stmt.value, ast.Name):
-                    returned_var = _ssa_get(stmt.value.id)
+                    name = stmt.value.id
+                    if name in latest:
+                        returned_var = _ssa_get(name)
+                    else:
+                        const_id = _ssa_new("return_value")
+                        ops.append({
+                            "id": const_id,
+                            "op": "CONST.value",
+                            "deps": [],
+                            "args": {"value": None},
+                        })
+                        returned_var = const_id
                 elif isinstance(stmt.value, (ast.Constant, ast.List, ast.Tuple, ast.Dict)):
-                    lit = _literal(stmt.value)
-                    const_id = _ssa_new("return_value")
-                    ops.append({
-                        "id": const_id,
-                        "op": "CONST.value",
-                        "deps": [],
-                        "args": {"value": lit},
-                    })
-                    returned_var = const_id
+                    try:
+                        lit = _literal(stmt.value)
+                    except DSLParseError:
+                        returned_var = _emit_assign_from_literal_or_pack("return_value", stmt.value)
+                    else:
+                        const_id = _ssa_new("return_value")
+                        ops.append({
+                            "id": const_id,
+                            "op": "CONST.value",
+                            "deps": [],
+                            "args": {"value": lit},
+                        })
+                        returned_var = const_id
                 else:
                     raise DSLParseError("return must return a variable name or literal")
                 return None
@@ -410,15 +613,53 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
                 saved_ctx = context_suffix
                 ctx_counts["loop"] += 1
                 context_suffix = f"loop{ctx_counts['loop']}"
+                loop_depth += 1
                 versions, latest = versions_body, latest_body
+                # Predefine loop target variables as items from iterator for dependency resolution
+                def _bind_loop_target(target: ast.AST):
+                    if isinstance(target, ast.Name):
+                        ssa_item = _ssa_new(target.id)
+                        ops.append({
+                            "id": ssa_item,
+                            "op": "ITER.item",
+                            "deps": [iter_id],
+                            "args": {"target": target.id},
+                        })
+                    elif isinstance(target, ast.Tuple):
+                        for elt in target.elts:
+                            if isinstance(elt, ast.Name):
+                                ssa_item = _ssa_new(elt.id)
+                                ops.append({
+                                    "id": ssa_item,
+                                    "op": "ITER.item",
+                                    "deps": [iter_id],
+                                    "args": {"target": elt.id},
+                                })
+                    # Other patterns are ignored for now
+
+                _bind_loop_target(stmt.target)
                 for inner in stmt.body:
                     _parse_stmt(inner)
+                loop_depth -= 1
                 versions_body, latest_body = versions, latest
                 versions, latest = saved_versions, saved_latest
                 context_suffix = saved_ctx
-                # Add iter dep to first op in body
+                # Add iter dep to first op in body if not already present
                 if len(ops) > body_ops_start:
-                    ops[body_ops_start]["deps"] = [*ops[body_ops_start].get("deps", []), iter_id]
+                    first = ops[body_ops_start]
+                    deps0 = first.get("deps", []) or []
+                    if iter_id not in deps0:
+                        first["deps"] = [*deps0, iter_id]
+                # Emit a summary foreach comp node depending on iterable value deps
+                iter_name_deps = _collect_value_deps(stmt.iter)
+                foreach_deps = [_ssa_get(n) for n in iter_name_deps]
+                ssa_foreach = _ssa_new("foreach")
+                ops.append({
+                    "id": ssa_foreach,
+                    "op": "COMP.foreach",
+                    "deps": foreach_deps,
+                    "args": {"target": t_label or ""},
+                })
                 # Loop-carried vars: only those existing pre-loop and reassigned in body
                 changed = {k for k in latest_body if pre_latest.get(k) != latest_body.get(k)}
                 carried = [k for k in changed if k in pre_latest]
@@ -442,14 +683,19 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
                 saved_ctx = context_suffix
                 ctx_counts["while"] += 1
                 context_suffix = f"while{ctx_counts['while']}"
+                loop_depth += 1
                 versions, latest = versions_body, latest_body
                 for inner in stmt.body:
                     _parse_stmt(inner)
+                loop_depth -= 1
                 versions_body, latest_body = versions, latest
                 versions, latest = saved_versions, saved_latest
                 context_suffix = saved_ctx
                 if len(ops) > body_ops_start:
-                    ops[body_ops_start]["deps"] = [*ops[body_ops_start].get("deps", []), cond_id]
+                    first = ops[body_ops_start]
+                    deps0 = first.get("deps", []) or []
+                    if cond_id not in deps0:
+                        first["deps"] = [*deps0, cond_id]
                 changed = {k for k in latest_body if pre_latest.get(k) != latest_body.get(k)}
                 carried = [k for k in changed if k in pre_latest]
                 for var in sorted(carried):
@@ -461,20 +707,67 @@ def parse(source: str, function_name: Optional[str] = None) -> Dict[str, Any]:
                         "args": {"var": var},
                     })
                 return None
-            elif isinstance(stmt, (ast.Pass,)):
+            elif isinstance(stmt, ast.Try):
+                pre_versions = dict(versions)
+                pre_latest = dict(latest)
+                for inner in stmt.body:
+                    _parse_stmt(inner)
+                # else block executes only if no exception
+                for inner in getattr(stmt, "orelse", []):
+                    _parse_stmt(inner)
+                post_versions = versions
+                post_latest = latest
+                for handler in getattr(stmt, "handlers", []):
+                    saved_ctx = context_suffix
+                    ctx_counts["except"] = ctx_counts.get("except", 0) + 1
+                    context_suffix = f"except{ctx_counts['except']}"
+                    saved_versions, saved_latest = versions, latest
+                    versions, latest = dict(pre_versions), dict(pre_latest)
+                    for inner in handler.body:
+                        _parse_stmt(inner)
+                    versions, latest = post_versions, post_latest
+                    context_suffix = saved_ctx
+                for inner in getattr(stmt, "finalbody", []):
+                    _parse_stmt(inner)
+                return None
+            elif isinstance(stmt, ast.Break):
+                ssa = _ssa_new("break")
+                ops.append({"id": ssa, "op": "CTRL.break", "deps": [], "args": {}})
+                return None
+            elif isinstance(stmt, (ast.Pass, ast.Continue)):
                 return None
             else:
                 raise DSLParseError("Only assignments, control flow, settings/output calls, and return are allowed in function body")
 
         # Parse body sequentially; still require a resulting output
-        for i, stmt in enumerate(fn.body):  # type: ignore[attr-defined]
+        body = list(getattr(fn, "body", []))  # type: ignore[attr-defined]
+        if body and isinstance(body[0], ast.Expr) and isinstance(getattr(body[0], "value", None), ast.Constant) and isinstance(getattr(body[0].value, "value", None), str):  # type: ignore[attr-defined]
+            body = body[1:]
+        for stmt in body:
             _parse_stmt(stmt)
 
+        # If no explicit return was encountered, emit a terminal break node so
+        # that the plan represents function completion.
+        break_id: Optional[str] = None
+        if returned_var is None:
+            break_id = _ssa_new("break")
+            ops.append({"id": break_id, "op": "CTRL.break", "deps": [], "args": {}})
+
+        # If no outputs were produced, synthesise a default return of `None` so
+        # that parsing succeeds for empty functions.
+        if returned_var is None and not outputs:
+            const_id = _ssa_new("return_value")
+            ops.append({
+                "id": const_id,
+                "op": "CONST.value",
+                "deps": [],
+                "args": {"value": None},
+            })
+            returned_var = const_id
+
         if not outputs:
-            if returned_var is not None:
-                outputs.append({"from": returned_var, "as": "return"})
-            else:
-                raise DSLParseError("At least one output() call required")
+            outputs.append({"from": returned_var if returned_var is not None else break_id, "as": "return"})
+
         if len(ops) > 2000:
             raise DSLParseError("Too many operations")
 
